"""API endpoints for WhispyWyser services."""
from __future__ import annotations

import logging
import os
from pathlib import Path
from typing import Optional

import uvicorn
from fastapi import FastAPI, HTTPException
from pydantic import BaseModel, Field

from .spell_checker import JamSpellChecker, SpellCheckRequest, SpellCheckResponse

_LOGGER = logging.getLogger(__name__)

class SpellCheckRequestModel(BaseModel):
    """Request model for spell checking."""
    text: str = Field(..., description="Text to check and correct")
    language: str = Field(default="en", description="Language code (default: en)")

class SpellCheckResponseModel(BaseModel):
    """Response model for spell checking."""
    original: str = Field(..., description="Original input text")
    corrected: str = Field(..., description="Corrected text")
    corrected_words: list[tuple[str, str]] = Field(
        default_factory=list,
        description="List of (original, corrected) word pairs"
    )

def create_app(
    spell_checker: Optional[JamSpellChecker] = None,
    model_path: Optional[Path] = None,
    alphabet_path: Optional[Path] = None,
) -> FastAPI:
    """Create and configure the FastAPI application."""
    app = FastAPI(
        title="WhispyWyser API",
        description="API for WhispyWyser services",
        version="0.1.0",
    )
    
    # Initialize the spell checker if not provided
    if spell_checker is None:
        spell_checker = JamSpellChecker.get_instance()
        if model_path is None:
            model_path = Path("models/jamspell.bin")
        if alphabet_path is None:
            alphabet_path = Path("models/alphabet.txt")
        
        try:
            spell_checker.load_model(model_path, alphabet_path)
        except Exception as e:
            _LOGGER.warning("Failed to load spell checker model: %s", e)
    
    @app.get("/health")
    async def health_check() -> dict:
        """Health check endpoint."""
        return {"status": "ok"}
    
    @app.post("/api/spell/check", response_model=SpellCheckResponseModel)
    async def check_spelling(request: SpellCheckRequestModel) -> SpellCheckResponseModel:
        """Check and correct spelling in the input text."""
        try:
            response = spell_checker.correct(
                text=request.text,
                language=request.language
            )
            
            return SpellCheckResponseModel(
                original=response.original,
                corrected=response.corrected,
                corrected_words=response.corrected_words
            )
        except Exception as e:
            _LOGGER.error("Error in spell checking: %s", e, exc_info=True)
            raise HTTPException(status_code=500, detail=str(e)) from e
    
    return app

def run_api_server(
    host: str = "0.0.0.0",
    port: int = 8000,
    model_path: Optional[Path] = None,
    alphabet_path: Optional[Path] = None,
    reload: bool = False
) -> None:
    """Run the API server."""
    app = create_app(model_path=model_path, alphabet_path=alphabet_path)
    
    uvicorn.run(
        app,
        host=host,
        port=port,
        log_level="info",
        reload=reload
    )



#!/usr/bin/env python3
import argparse
import asyncio
import logging
import sys
import os
import subprocess
import shutil
import json
from functools import partial
from typing import Any

import faster_whisper
from wyoming.info import AsrModel, AsrProgram, Attribution, Info
from wyoming.server import AsyncServer

from . import __version__
from .handler import FasterWhisperEventHandler

_LOGGER = logging.getLogger(__name__)


async def main() -> None:
    """Main entry point."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--model",
        required=True,
        help="Name of faster-whisper model to use",
    )
    parser.add_argument(
        "--uri",
        required=True,
        help="unix:// or tcp://",
    )
    parser.add_argument(
        "--data-dir",
        required=True,
        action="append",
        help="Data directory to check for downloaded models",
    )
    parser.add_argument(
        "--model-dir",
        required=True,
        help="Directory to download/load whisper models",
    )
    parser.add_argument(
        "--device",
        default="cpu",
        choices=["cpu", "cuda", "auto"],
        help="Device to use for inference (default: cpu, cuda, auto)",
    )
    parser.add_argument(
        "--cpu-threads",
        type=int,
        default=0,
        help="Number of CPU threads to use for inference in case of CPU device (default: 0, which means all available threads)",
    )
    parser.add_argument(
        "--language",
        help="Default language to set for transcription",
    )
    parser.add_argument(
        "--compute-type",
        default="default",
        help="Compute type (default, auto, int8, int_float32, int8_float16, int8_bfloat16, int16, float16, float32, bfloat16)",
    )
    parser.add_argument(
        "--beam-size",
        type=int,
        default=5,
        help="Size of beam during decoding",
    )
    parser.add_argument(
        "--initial-prompt",
        help="Optional text to provide as a prompt for the first window",
    )
    parser.add_argument(
        "--local-files-only",
        action="store_true",
        help="Don't check HuggingFace hub for updates every time",
    )
    parser.add_argument(
        "--model-type",
        default="ct2",
        choices=["ct2", "distil", "transformers", "convert"],
        help="model type to use (ct2[faster-whisper], distil, transformers, convert[convert transformers model to CTranslate2 format]). default: ct2",
    )
    #
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Log DEBUG messages",
    )
    parser.add_argument(
        "--log-format",
        default=logging.BASIC_FORMAT,
        help="Format for log messages",
    )
    parser.add_argument(
        "--version",
        action="version",
        version=__version__,
        help="Print version and exit",
    )
    args = parser.parse_args()

    logging.basicConfig(
        level=logging.DEBUG if args.debug else logging.INFO, format=args.log_format
    )
    _LOGGER.debug(
        "args:\n%s", json.dumps(vars(args), indent=4, sort_keys=True)
    )

    # Resolve model name
    model_name = args.model

    if args.language == "auto":
        # Whisper does not understand "auto"
        args.language = None

    hf_token = None
    if os.getenv("HUGGINGFACE_HUB_TOKEN"):
        hf_token = os.getenv("HUGGINGFACE_HUB_TOKEN")
    elif os.getenv("HF_TOKEN"):
        hf_token = os.getenv("HF_TOKEN")


    wyoming_info = Info(
        asr=[
            AsrProgram(
                name="whispywiser",
                description="Custom Faster Whisper transcription with CTranslate2",
                attribution=Attribution(
                    name="Guillaume Klein",
                    url="https://github.com/guillaumekln/faster-whisper/",
                ),
                installed=True,
                version=__version__,
                models=[
                    AsrModel(
                        name=model_name,
                        description=model_name,
                        attribution=Attribution(
                            name="Systran",
                            url="https://huggingface.co/Systran",
                        ),
                        installed=True,
                        languages=faster_whisper.tokenizer._LANGUAGE_CODES,  # pylint: disable=protected-access
                        version=faster_whisper.__version__,
                    )
                ],
            )
        ],
    )

    _LOGGER.debug(
        "wyoming_info:\n%s", json.dumps(vars(wyoming_info), indent=4, sort_keys=True)
    )

    # Load model
    _LOGGER.debug("Loading %s", args.model)
    whisper_model: Any = None
    model_type = args.model_type.lower() or "ct2"
    if model_type == "convert":
        # Convert model
        try:
            converter_path = shutil.which("ct2-transformers-converter")
            if converter_path is None:
                raise FileNotFoundError("ct2-transformers-converter not found in PATH.")
            logging.info(f"Found converter at: {converter_path}")

            model_path = os.path.join(args.model_dir, model_name)
            model_conversion_path = os.path.join(args.model_dir, model_name + "-converted")
            DONE_FLAG = os.path.join(model_conversion_path, ".done")

            if os.path.exists(DONE_FLAG):
                print("Model already converted. Skipping conversion.")
                converted = True
                model_name = model_name + "-converted"
                subprocess_task_converted = True

            if not converted:
                if "int8" not in args.compute_type:
                    quantization = "int8"
                else:
                    quantization = "float16"

                result = subprocess.run([
                    "ct2-transformers-converter",
                    "--model", model_path,
                    "--output_dir", model_conversion_path,
                    "--quantization", quantization,
                    "--copy_files", "tokenizer.json", "preprocessor_config.json"
                ], check=True, capture_output=True, text=True)
                logging.info("Conversion successful:\n%s", result.stdout)
                open(DONE_FLAG, "w").close()
                model_name = model_name + "-converted"
                subprocess_task_converted = True

        except subprocess.CalledProcessError as e:
            logging.error("Conversion failed:\n%s", e.stderr)
            sys.exit("Conversion process failed. Check logs.")
        except FileNotFoundError as e:
            logging.error(f"{e}")
            sys.exit("Exiting: Required tool not found.")
        except Exception as e:
            logging.error(f"Unexpected error: {e}")
            sys.exit("Exiting: Unexpected error.")

    if model_type == "transformers":
        # Use HuggingFace transformers
        from .transformers_whisper import TransformersWhisperModel

        whisper_model = TransformersWhisperModel(
            args.model, args.model_dir, args.local_files_only
        )
    elif model_type == "ct2" or model_type == "distil" or subprocess_task_converted:
        # Use faster-whisper
        whisper_model = faster_whisper.WhisperModel(
            args.model,
            download_root=args.model_dir,
            device=args.device.lower(),
            compute_type=args.compute_type,
            cpu_threads=args.cpu_threads,
            local_files_only=args.local_files_only,
            use_auth_token=hf_token,
        )

    server = AsyncServer.from_uri(args.uri)
    _LOGGER.info("Ready")
    model_lock = asyncio.Lock()

    if model_type == "transformers":
        # Use HuggingFace transformers
        from .transformers_whisper import (
            TransformersWhisperEventHandler,
            TransformersWhisperModel,
        )

        assert isinstance(whisper_model, TransformersWhisperModel)

        # TODO: initial prompt
        await server.run(
            partial(
                TransformersWhisperEventHandler,
                wyoming_info,
                args.language,
                args.beam_size,
                whisper_model,
                model_lock,
            )
        )
    elif model_type == "ct2" or model_type == "distil" or subprocess_task_converted:
        # Use faster-whisper
        assert isinstance(whisper_model, faster_whisper.WhisperModel)
        await server.run(
            partial(
                FasterWhisperEventHandler,
                wyoming_info,
                args,
                whisper_model,
                model_lock,
                initial_prompt=args.initial_prompt,
            )
        )


# -----------------------------------------------------------------------------


def run() -> None:
    asyncio.run(main())


if __name__ == "__main__":
    try:
        run()
    except KeyboardInterrupt:
        pass



"""Home Assistant WebSocket and REST API client."""
from __future__ import annotations

import asyncio
import logging
from typing import Any, Dict, List, Optional, Callable, Awaitable

import aiohttp

from .models import Area, Entity, Things

_LOGGER = logging.getLogger(__name__)

RGB_MODES = {"hs", "rgb", "rgbw", "rgbww", "xy"}
BRIGHTNESS_MODES = RGB_MODES | {"brightness", "white"}
FAN_SET_SPEED = 1
COVER_SET_POSITION = 4
MEDIA_PLAYER_PAUSE = 1
MEDIA_PLAYER_VOLUME_SET = 4
MEDIA_PLAYER_NEXT_TRACK = 32

class HomeAssistantError(Exception):
    """Base exception for Home Assistant errors."""
    pass

def validate_ha_token(token: Optional[str]) -> str:
    """Validate Home Assistant token.
    
    Args:
        token: Token to validate (can be None)
        
    Returns:
        Validated token
        
    Raises:
        HomeAssistantError: If token is invalid or not provided
    """
    if not token:
        raise HomeAssistantError(
            "No Home Assistant token provided. "
            "Please set the HA_TOKEN environment variable."
        )
        
    if not isinstance(token, str) or len(token.strip()) < 10:  # Basic validation
        raise HomeAssistantError("Invalid Home Assistant token format")
        
    return token.strip()


class HomeAssistantClient:
    """Client for interacting with Home Assistant's API."""
    
    def __init__(
        self,
        url: str,
        token: str,
        session: Optional[aiohttp.ClientSession] = None,
    ) -> None:
        """Initialize the Home Assistant client.
        
        Args:
            url: Base URL of the Home Assistant instance (e.g., 'http://homeassistant:8123')
            token: Long-lived access token
            session: Optional aiohttp ClientSession
            
        Raises:
            HomeAssistantError: If token is invalid
        """
        self._base_url = url.rstrip('/')
        self._token = validate_ha_token(token)
        self._headers = {
            'Authorization': f'Bearer {self._token}',
            'Content-Type': 'application/json',
        }
        self._session = session or aiohttp.ClientSession()
        self._websocket = None
        self._message_id = 1
        self._listeners = {}
        self._event_listeners = {}
        
    async def __aenter__(self):
        await self.connect()
        return self
        
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()
        
    async def connect(self) -> None:
        """Connect to Home Assistant API and set up WebSocket."""
        if self._session.closed:
            self._session = aiohttp.ClientSession()
            
        await self._setup_websocket()
            
    async def close(self) -> None:
        """Close the connection and clean up resources."""
        if self._websocket is not None:
            await self._websocket.close()
            self._websocket = None
            
        if not self._session.closed:
            await self._session.close()
            
    async def _setup_websocket(self) -> None:
        """Set up WebSocket connection to Home Assistant."""
        ws_url = self._base_url.replace('http', 'ws') + '/api/websocket'
        
        try:
            self._websocket = await self._session.ws_connect(ws_url, heartbeat=55)
            auth_required = await self._websocket.receive_json()
            
            if auth_required['type'] != 'auth_required':
                raise HomeAssistantError('Unexpected message during WebSocket auth')
                
            await self._websocket.send_json({
                'type': 'auth',
                'access_token': self._token
            })
            
            auth_result = await self._websocket.receive_json()
            if auth_result['type'] != 'auth_ok':
                raise HomeAssistantError(f'Authentication failed: {auth_result}')
                
            # Start listening for events
            asyncio.create_task(self._websocket_listener())
            
        except Exception as err:
            _LOGGER.error('WebSocket connection failed: %s', err)
            raise HomeAssistantError(f'WebSocket connection failed: {err}') from err
            
    async def _websocket_listener(self) -> None:
        """Listen for WebSocket messages and dispatch to handlers."""
        try:
            async for msg in self._websocket:
                if msg.type == aiohttp.WSMsgType.TEXT:
                    data = msg.json()
                    await self._handle_websocket_message(data)
                elif msg.type == aiohttp.WSMsgType.CLOSED:
                    _LOGGER.warning('WebSocket connection closed')
                    break
                elif msg.type == aiohttp.WSMsgType.ERROR:
                    _LOGGER.error('WebSocket error: %s', self._websocket.exception())
                    break
                    
        except Exception as err:
            _LOGGER.error('WebSocket listener error: %s', err)
        finally:
            self._websocket = None
            
    async def _handle_websocket_message(self, data: Dict[str, Any]) -> None:
        """Handle incoming WebSocket messages."""
        if 'id' in data and data['id'] in self._listeners:
            future = self._listeners.pop(data['id'])
            if not future.done():
                if data.get('success', True):
                    future.set_result(data.get('result'))
                else:
                    future.set_exception(HomeAssistantError(data.get('error', {}).get('message', 'Unknown error')))
                    
        elif data.get('type') == 'event':
            event_type = data.get('event', {}).get('event_type')
            if event_type in self._event_listeners:
                for callback in self._event_listeners[event_type]:
                    try:
                        await callback(data['event'])
                    except Exception as err:
                        _LOGGER.error('Error in event handler: %s', err)
                        
    async def _make_request(self, method: str, endpoint: str, **kwargs) -> Any:
        """Make an HTTP request to the Home Assistant API."""
        url = f"{self._base_url}/api/{endpoint.lstrip('/')}"
        
        if 'headers' not in kwargs:
            kwargs['headers'] = {}
        kwargs['headers'].update(self._headers)
        
        try:
            async with self._session.request(method, url, **kwargs) as resp:
                resp.raise_for_status()
                if resp.status == 204:  # No content
                    return None
                return await resp.json()
        except aiohttp.ClientError as err:
            _LOGGER.error('Request failed: %s', err)
            raise HomeAssistantError(f'Request failed: {err}') from err
            
    async def get_entities(self, debug_logger = None) -> List[Dict[str, Any]]:
        """Get all entities from Home Assistant.
        
        Args:
            debug_logger: Optional debug logger instance for logging discovered entities
            
        Returns:
            List of entity states
        """
        entities = await self._make_request('GET', 'states')
        
        # Log entities if debug logger is provided
        if debug_logger and hasattr(debug_logger, 'log_entities'):
            try:
                # Create a Things object with the discovered entities
                things = await self.get_hass_info()
                debug_logger.log_entities(things)
            except Exception as e:
                _LOGGER.warning("Failed to log entities for debugging: %s", e)
                
        return entities
        
    async def get_areas(self) -> List[Dict[str, Any]]:
        """Get all areas from Home Assistant."""
        return await self._make_request('GET', 'config/area_registry')
        
    async def get_entity_registry(self) -> List[Dict[str, Any]]:
        """Get the entity registry from Home Assistant."""
        return await self._make_request('GET', 'config/entity_registry')
        
    async def get_services(self) -> Dict[str, Any]:
        """Get all available services from Home Assistant."""
        return await self._make_request('GET', 'services')
        
    async def call_service(
        self,
        domain: str,
        service: str,
        service_data: Optional[Dict[str, Any]] = None,
        target: Optional[Dict[str, Any]] = None,
    ) -> Any:
        """Call a service in Home Assistant."""
        data = {}
        if service_data:
            data["service_data"] = service_data
        if target:
            data["target"] = target
            
        return await self._make_request(
            'POST',
            f'services/{domain}/{service}',
            json=data
        )

    def subscribe_events(
        self,
        event_type: str,
        callback: Callable[[Dict[str, Any]], Awaitable[None]],
    ) -> Callable[[], None]:
        """Subscribe to Home Assistant events.
        
        Args:
            event_type: Event type to subscribe to
            callback: Async callback to call when event is received
            
        Returns:
            Function to unsubscribe
        """
        if event_type not in self._event_listeners:
            self._event_listeners[event_type] = []
            
        self._event_listeners[event_type].append(callback)
        
        def unsubscribe():
            if event_type in self._event_listeners and callback in self._event_listeners[event_type]:
                self._event_listeners[event_type].remove(callback)
                
        return unsubscribe

async def get_hass_info(
    url: str,
    token: str,
    session: Optional[aiohttp.ClientSession] = None,
    debug_logger: Optional[Any] = None,
) -> Things:
    """Get exposed entities and areas from Home Assistant."""
    things = Things()
    
    async with HomeAssistantClient(url, token, session) as client:
        # Get all entities and their states
        entities = await client.get_entities(debug_logger=debug_logger)
        entity_registry = await client.get_entity_registry()
        areas = await client.get_areas()
        
        # Process areas
        area_map = {area["area_id"]: area for area in areas}
        
        for area in areas:
            area_names = []
            if "name" in area and area["name"]:
                area_names.append(area["name"])
                
            if area_names:
                things.areas.append(
                    Area(area_id=area["area_id"], names=area_names)
                )
        
        # Process entities
        entity_registry_map = {
            entry["entity_id"]: entry for entry in entity_registry
        }
        
        for entity_state in entities:
            entity_id = entity_state["entity_id"]
            domain = entity_id.split(".", 1)[0]
            
            # Skip entities from ignored domains
            if domain not in {"light", "switch", "media_player", "fan", "cover", "climate"
            }:
                continue
                
            # Get names from entity registry
            names = []
            registry_entry = entity_registry_map.get(entity_id, {})
            
            # Add friendly name if available
            name = registry_entry.get("name") or entity_state.get("attributes", {}).get("friendly_name")
            if name:
                names.append(name)
                
            # Add area name if available
            area_id = registry_entry.get("area_id")
            if area_id and area_id in area_map:
                area = area_map[area_id]
                if "name" in area and area["name"]:
                    names.append(f"{area['name']} {name}" if name else area["name"])
            
            if not names:
                continue
                
            # Create entity with basic info
            entity = Entity(
                entity_id=entity_id,
                names=names,
                domain=domain
            )
            
            # Add domain-specific features
            attributes = entity_state.get("attributes", {})
            
            if domain == "light":
                supported_features = attributes.get("supported_features", 0)
                entity.light_supports_color = bool(supported_features & 16)  # SUPPORT_COLOR
                entity.light_supports_brightness = bool(supported_features & 1)  # SUPPORT_BRIGHTNESS
                
            elif domain == "fan":
                supported_features = attributes.get("supported_features", 0)
                entity.fan_supports_speed = bool(supported_features & FAN_SET_SPEED)
                
            elif domain == "cover":
                supported_features = attributes.get("supported_features", 0)
                entity.cover_supports_position = bool(supported_features & COVER_SET_POSITION)
                
            elif domain == "media_player":
                supported_features = attributes.get("supported_features", 0)
                entity.media_player_supports_pause = bool(supported_features & MEDIA_PLAYER_PAUSE)
                entity.media_player_supports_volume_set = bool(supported_features & MEDIA_PLAYER_VOLUME_SET)
                entity.media_player_supports_next_track = bool(supported_features & MEDIA_PLAYER_NEXT_TRACK)
            
            things.entities.append(entity)
    
    return things



"""Spell checking functionality using JamSpell."""
from __future__ import annotations

import logging
import os
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, Tuple, Dict, Any, ClassVar

import jamspell

_LOGGER = logging.getLogger(__name__)

@dataclass
class SpellCheckResponse:
    """Response from spell checking."""
    original: str
    corrected: str
    corrected_words: List[Tuple[str, str]]
    
    @property
    def was_corrected(self) -> bool:
        """Return True if any corrections were made."""
        return self.original != self.corrected

class JamSpellChecker:
    """Wrapper around JamSpell for spell checking."""
    
    _instance: ClassVar[Optional[JamSpellChecker]] = None
    
    def __init__(self):
        """Initialize the spell checker."""
        self._model: Optional[jamspell.TSpellCorrector] = None
        self._loaded_models: Dict[str, str] = {}  # language: model_path
    
    @classmethod
    def get_instance(cls) -> JamSpellChecker:
        """Get or create the singleton instance."""
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance
    
    def load_model(self, model_path: Path, alphabet_path: Path) -> None:
        """Load a JamSpell model.
        
        Args:
            model_path: Path to the .bin model file
            alphabet_path: Path to the alphabet file
        """
        if not model_path.exists() or not alphabet_path.exists():
            raise FileNotFoundError(
                f"Model files not found. Model: {model_path}, Alphabet: {alphabet_path}"
            )
        
        _LOGGER.info("Loading JamSpell model from %s", model_path)
        
        self._model = jamspell.TSpellCorrector()
        
        if not self._model.LoadLangModel(str(model_path)):
            raise RuntimeError(f"Failed to load JamSpell model from {model_path}")
        
        _LOGGER.info("Successfully loaded JamSpell model")
    
    def correct(self, text: str, language: str = "en") -> SpellCheckResponse:
        """Correct spelling in the input text.
        
        Args:
            text: Input text to correct
            language: Language code (default: en)
            
        Returns:
            SpellCheckResponse with original and corrected text
        """
        if not text.strip():
            return SpellCheckResponse(text, text, [])
            
        if self._model is None:
            _LOGGER.warning("No JamSpell model loaded, returning original text")
            return SpellCheckResponse(text, text, [])
        
        try:
            corrected_text = self._model.FixFragment(text)
            
            # Get the list of corrected words
            corrected_words: List[Tuple[str, str]] = []
            words = text.split()
            corrected_words_list = corrected_text.split()
            
            for orig_word, corrected_word in zip(words, corrected_words_list):
                if orig_word != corrected_word:
                    corrected_words.append((orig_word, corrected_word))
            
            return SpellCheckResponse(
                original=text,
                corrected=corrected_text,
                corrected_words=corrected_words
            )
            
        except Exception as e:
            _LOGGER.error("Error in spell checking: %s", e, exc_info=True)
            # Return original text on error
            return SpellCheckResponse(text, text, [])
